#coding=utf-8
'''這是一個HTML編輯器，使用QT4，WebKit建立'''

import os ,time ,platform, fnmatch, shutil, functools

from PyQt4.QtGui import *
from PyQt4.QtCore import *
from PyQt4.QtWebKit import *

import lonote.path as path
import lonote.gadget as gadget


class LoWebView(QWebView):
    '''把Context Menu關掉的QWebView'''

    imageUrlsDrop = pyqtSignal(list) #image連結被拖放之後的結果

    def contextMenuEvent(self, ev):
        '''重設右鍵選單'''
        clipboard = QApplication.clipboard()
        menu = QMenu() #總清單
        #建立action
        #貼上純文字
        act = QAction(_('Paste plain text'), menu)
        act.triggered.connect(functools.partial(self.page().triggerAction, QWebPage.PasteAndMatchStyle))
        menu.addAction(act)
        #複製純文字
        act = QAction(_('Copy plain text'), menu)
        def copyPlainText():
            '''如果selectText中有東西，複製。否則為空操作。'''
            if self.page().selectedText():
                clipboard.setText(self.page().selectedText())
        act.triggered.connect(copyPlainText)
        menu.addAction(act)
        menu.addSeparator()
        ##undo
        #act = QAction(_('Undo'), menu)
        #act.triggered.connect(functools.partial(self.page().triggerAction, QWebPage.Undo))
        #menu.addAction(act)
        ##redo
        #act = QAction(_('Redo'), menu)
        #act.triggered.connect(functools.partial(self.page().triggerAction, QWebPage.Redo))
        #menu.addAction(act)
        #menu.addSeparator()
        #複製連結到剪貼簿
        act = QAction(_('Copy link to clipboard'), menu)
        act.triggered.connect(functools.partial(self.page().triggerAction, QWebPage.CopyLinkToClipboard))
        menu.addAction(act)

        menu.exec_(QCursor.pos())

    def dragMoveEvent(self, event):
        super().dragMoveEvent(event)

    def dropEvent(self, event):
        '''用拖拉來插入圖檔'''
        data = event.mimeData()
        #如果從外面拖進來且含有url，則執行插入圖檔任務
        if event.source() is None and data.hasUrls():
            #先解析urls
            urls = [url.toLocalFile() for url in data.urls()
                    if url.scheme() == 'file' and
                    gadget.isSupportImage(url.toLocalFile())]
            #將urls扔去插入圖片用
            if urls != []:
                print(urls)
                #先變更游標位置
                JS = gadget.JS_setCaretPosByMousePos.replace(
                        '{$POSX}', str(event.pos().x()))
                JS = JS.replace('{$POSY}', str(event.pos().y()))
                self.page().mainFrame().evaluateJavaScript(JS)
                #執行插入指令
                self.imageUrlsDrop.emit(urls)
                return
        super().dropEvent(event)

class LoEditor(QWidget):
    '''這是一個Qt Html編輯器'''

    SIGNALattachAdd = pyqtSignal()
    lowikiClickedSignal = pyqtSignal(int)

#初始化

    def __init__(self, parent = None):
        super(QWidget, self).__init__(parent)

        self.noteArea = LoWebView()
        self.editToolbar = QToolBar()
        self.findbar = QToolBar()
        self.findEdit = QLineEdit()
        #self.replaceEdit = QLineEdit()

        #設定要給插入時間和插入日期的格式
        self.DATE_FORMAT = '%Y %B %d'
        self.TIME_FORMAT = '%p %H:%M'
        self.IMG_FORMAT = '<p>{$IMAGE}</p><br>'
        self.IMG_FORMAT_WIDTH = 500
        self.IMG_FORMAT_RESIZE_ENABLE = True
        self.IMG_FORMAT_INSERTRESIZE_ENABLE = True
        self.IMG_FORMAT_INSERTRESIZE_QUALITY = -1 #範圍為-1~100，100品質最好，0最差，-1是預設
        self.FONT_ST = 'wqy-microhei' #編輯器標準字型
        self.TEXT_SIZE_MULTIPLIER = 1.0 #文字放大率

        self.db = None #供lowiki使用的資料庫，沒有時相關操作一律為空操作。

        self.cache = {} #用id索引(page,embpath,attpath)的字典
        self.pageid = None #當前顯示中的pageid
        self.pagedir = None #當前的notepath
        self.pageemb = None #當前的embeddedpath
        self.pageatt = None #當前的attachedpath

        self._initLayout()
        self._initEditFun()
        self._initTool()
        self._initNoteArea()

    def _initNoteArea(self):
        self.noteArea.setRenderHint(QPainter.Antialiasing, True)
        self.noteArea.setRenderHint(QPainter.TextAntialiasing, True)
        self.noteArea.setRenderHint(QPainter.SmoothPixmapTransform, True)
        self.noteArea.setRenderHint(QPainter.HighQualityAntialiasing, True)
        #圖片連結被拖入後進行複製
        self.noteArea.imageUrlsDrop.connect(self._inimg)

    def _initLayout(self):
        '''初始化noteArea和editToolbar等Layout'''
        #======findbar的設定======
        label = QLabel(_('Find: '))
        self.findEdit = QLineEdit()
        self.findEdit.setMaximumWidth(350)
        self.findEdit.setStatusTip(_('This is a search bar. Please input some word you want to find.'))
        #findbar快速鍵召喚＼反召喚
        self.shortcut_findbarShow = QShortcut(QKeySequence("Ctrl+F"), self.noteArea)
        self.shortcut_findbarShow.activated.connect(self.findbarShow)
        self.shortcut_findbarHide = QShortcut(QKeySequence("ESC"), self.findbar)
        self.shortcut_findbarHide.activated.connect(self.findbarHide)
        #關閉按鈕
        closeButton = QPushButton(QIcon(path.icons.close), '', self.findbar)
        closeButton.setStatusTip(_('Close The Search Bar (Esc)'))
        closeButton.pressed.connect(self.findbarHide)
        closeButton.setFlat(True)
        #下一個按鈕
        nextButton = QPushButton(QIcon(path.icons.yes), _('Next＞'), self.findbar)
        nextButton.setStatusTip(_('Search the next match word. (Enter)'))
        nextButton.pressed.connect(self.findNext)
        #前一個
        beforeButton = QPushButton(QIcon(path.icons.before), _('＜Before'), self.findbar)
        beforeButton.setStatusTip(_('Search the match word before here.'))
        beforeButton.pressed.connect(self.findBefore)
        #findEdit connect
        self.findEdit.textChanged.connect(self.findAll)
        self.findEdit.returnPressed.connect(self.findNext)
        #鋪設findbarLayout
        self.findbar.addWidget(closeButton)
        self.findbar.addWidget(label)
        self.findbar.addWidget(self.findEdit)
        self.findbar.addWidget(beforeButton)
        self.findbar.addWidget(nextButton)
        self.findbar.hide()

        #======終於把findbar設定好了！======

        #Layout設定（由小到大）
        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self.noteArea)
        mainLayout.addWidget(self.findbar)
        mainLayout.addWidget(self.editToolbar)

        self.setLayout(mainLayout)

    def _initEditFun(self):
        '''設定所有編輯按鈕、選單、快捷鍵的功能、圖示與說明，但不包括安排顯示位置'''
        #黑體字:bold
        self.bold = QAction(QIcon(path.icons.bold), _('Bold'), self)
        self.bold.setShortcut('Ctrl+B')
        self.bold.setStatusTip(_('toggle bold (Ctrl+B)'))
        self.connect(self.bold, SIGNAL('triggered()'), SLOT('_bold()'))
        #斜體字:italic
        self.italic = QAction(QIcon(path.icons.italic), _('Italic'), self)
        self.italic.setShortcut('Ctrl+I')
        self.italic.setStatusTip(_('toggle italic (Ctrl+I)'))
        self.connect(self.italic, SIGNAL('triggered()'), SLOT('_italic()'))
        #底線字:underline
        self.underline = QAction(QIcon(path.icons.underline), _('Underline'), self)
        self.underline.setShortcut('Ctrl+U')
        self.underline.setStatusTip(_('toggle underline (Ctrl+U)'))
        self.connect(self.underline, SIGNAL('triggered()'), SLOT('_underline()'))
        #畫線字:strike
        self.strike = QAction(QIcon(path.icons.strike), _('Strike'), self)
        self.strike.setShortcut('Ctrl+S')
        self.strike.setStatusTip(_('toggle strike (Ctrl+S)'))
        self.strike.triggered.connect(self._strike)
        #上標:superscript
        self.superscript = QAction(QIcon(path.icons.sup), _('Superscript'), self)
        self.superscript.setStatusTip(_('toggle superscript (Alt+K)'))
        self.superscript.setShortcut('Alt+K')
        self.connect(self.superscript, SIGNAL('triggered()'), SLOT('_superscript()'))
        #下標:subscript
        self.subscript = QAction(QIcon(path.icons.sub), _('Subscript'), self)
        self.subscript.setStatusTip(_('toggle subscript (Alt+J)'))
        self.subscript.setShortcut('Alt+J')
        self.connect(self.subscript, SIGNAL('triggered()'), SLOT('_subscript()'))
        #靠左對齊:justifyLeft
        self.justifyLeft = QAction(QIcon(path.icons.left), _('Left'), self)
        self.justifyLeft.setStatusTip(_('Align to left (Alt+,)'))
        self.justifyLeft.setShortcut('Alt+,')
        self.connect(self.justifyLeft, SIGNAL('triggered()'), SLOT('_justifyLeft()'))
        #置中對齊:justifyCenter
        self.justifyCenter = QAction(QIcon(path.icons.center), _('Center'), self)
        self.justifyCenter.setStatusTip(_('Align to center (Alt+M)'))
        self.justifyCenter.setShortcut('Alt+M')
        self.connect(self.justifyCenter, SIGNAL('triggered()'), SLOT('_justifyCenter()'))
        #靠右對齊:justifyRight
        self.justifyRight = QAction(QIcon(path.icons.right), _('Right'), self)
        self.justifyRight.setStatusTip(_('Align to right (Alt+.)'))
        self.justifyRight.setShortcut('Alt+.')
        self.connect(self.justifyRight, SIGNAL('triggered()'), SLOT('_justifyRight()'))
        #縮排:indent
        self.indent = QAction(QIcon(path.icons.indent), _('Indent'), self)
        self.indent.setStatusTip(_('indent (Alt+])'))
        self.indent.setShortcut('Alt+]')
        self.connect(self.indent, SIGNAL('triggered()'), SLOT('_indent()'))
        #反縮排:outdent
        self.outdent = QAction(QIcon(path.icons.outdent), _('Outdent'), self)
        self.outdent.setStatusTip(_('outdent (Alt+[)'))
        self.outdent.setShortcut('Alt+[')
        self.connect(self.outdent, SIGNAL('triggered()'), SLOT('_outdent()'))
        #有序列表:ordlist
        self.ordlist = QAction(QIcon(path.icons.ordlist), _('Ordered list'), self)
        self.ordlist.setStatusTip(_('ordered list (Ctrl+O)'))
        self.ordlist.setShortcut('Ctrl+O')
        self.connect(self.ordlist, SIGNAL('triggered()'), SLOT('_ordlist()'))
        #無序列表:bullist
        self.bullist = QAction(QIcon(path.icons.bullist), _('Bullet list'), self)
        self.bullist.setStatusTip(_('bullet list (Ctrl+P)'))
        self.bullist.setShortcut('Ctrl+P')
        self.connect(self.bullist, SIGNAL('triggered()'), SLOT('_bullist()'))
        #前景色:fgcolor
        self.fgcolor = QAction(QIcon(path.icons.forecol), _('Text color'), self)
        self.fgcolor.setStatusTip(_('Select a text color. (Ctrl+E)'))
        self.fgcolor.setShortcut('Ctrl+E')
        self.fgcolor.triggered.connect(self._fgcolor)
        #背景色:bgcolor
        self.bgcolor = QAction(QIcon(path.icons.bgcol), _('Background color'), self)
        self.bgcolor.setStatusTip(_('Select a background color. (Ctrl+D)'))
        self.bgcolor.setShortcut('Ctrl+D')
        self.bgcolor.triggered.connect(self._bgcolor)
        #標籤
        self.headingBox = QComboBox(self)
        self.headingBox.addItems([_('--Tags--'),_('Paragraph'),_('Heading1'),_('Heading2'),_('Heading3'),_('Heading4'),_('Heading5'),_('Heading6'),_('Blockquote'),_('code'),_('address')])
        self.headingBox.setStatusTip(_('Assign line format. (Alt+` ~ Alt+9)'))
        self.headingBox.activated[int].connect(self._headingChanged)
        #字體大小
        self.fontsize = QComboBox(self)
        self.fontsize.addItems([_('Fontsize'),'１','２',_('Standard'),'４','５','６','７'])
        self.fontsize.setStatusTip(_('Assign font size.'))
        self.fontsize.activated[int].connect(self._fontsizeChanged)
        #插入圖片:inimg
        self.inimg = QAction(QIcon(path.icons.inimg), _('Insert images'), self)
        self.inimg.setStatusTip(_('Insert images on your cursor position. (Alt+I)'))
        self.inimg.setShortcut('Alt+I')
        self.inimg.triggered.connect(self._inimg)
        #加入超聯結:link
        self.link = QAction(QIcon(path.icons.link), _('Hyperlink'), self)
        self.link.setStatusTip(_('Convert to hyperlink (Ctrl+L)    ☆Hint: you can use "URL???TEXT" format.'))
        self.link.setShortcut('Ctrl+L')
        self.link.triggered.connect(self._link)
        #移除超聯結:unlink
        self.unlink = QAction(QIcon(path.icons.unlink), _('unlink'), self)
        self.unlink.setStatusTip(_('Remove selected hyperlink (Ctrl+Shift+K)'))
        self.unlink.setShortcut('Ctrl+Shift+K')
        self.unlink.triggered.connect(self._unlink)
        #去除格式:noformat
        self.noformat = QAction(QIcon(path.icons.noformat), _('Remove Format'), self)
        self.noformat.setStatusTip(_('Remove text format (Ctrl+K)'))
        self.noformat.setShortcut('Ctrl+K')
        self.noformat.triggered.connect(self._noformat)
        #插入附件:
        self.inatt = QAction(QIcon(path.icons.inatt), _('Add attachment'), self)
        self.inatt.setStatusTip(_('Add attachment in this page. (Ctrl+Alt+J)'))
        self.inatt.setShortcut('Ctrl+Alt+J')
        self.inatt.triggered.connect(self._inatt)
        #插入HTML原始碼：
        self.inhtml = QAction(QIcon(path.icons.inhtml), _('Insert HTML source'), self)
        self.inhtml.setStatusTip(_('Insert a HTML source code on your cursor position. (Ctrl+Alt+I)'))
        self.inhtml.setShortcut('Ctrl+Alt+I')
        self.inhtml.triggered.connect(self._inhtml)
        #插入日期
        self.indate = QAction(QIcon(path.icons.indate), _('Insert Date'), self)
        self.indate.setStatusTip(_('Insert current date on your cursor position (Alt+D)'))
        self.indate.setShortcut('Alt+D')
        self.indate.triggered.connect(self._indate)
        #插入時間
        self.intime = QAction(QIcon(path.icons.intime), _('Insert time'), self)
        self.intime.setStatusTip(_('Insert current time on your cursor position (Alt+T)'))
        self.intime.setShortcut('Alt+T')
        self.intime.triggered.connect(self._intime)
        #插入快速模板
        self.intemplate = QAction(QIcon(path.icons.intemplate), _('Quick template'), self)
        self.intemplate.setStatusTip(_('Insert a quick template on current cursor position (Alt+Q)    ☆Hint: The quick templates be defined in "{0}"').format(path.templateDir))
        self.intemplate.setShortcut('Alt+Q')
        self.intemplate.triggered.connect(self._intemplate)

        #連結被敲擊後的反應
        self.noteArea.linkClicked[QUrl].connect(self._linkClicked)

        #快捷鍵
        self.shortcut_P = QShortcut(QKeySequence("Alt+`"), self.noteArea)
        self.shortcut_P.activated.connect(self.fbP)
        self.shortcut_P2 = QShortcut(QKeySequence("Alt+0"), self.noteArea)
        self.shortcut_P2.activated.connect(self.fbP)

        self.shortcut_H1 = QShortcut(QKeySequence("Alt+1"), self.noteArea)
        self.shortcut_H1.activated.connect(self.fbH1)
        self.shortcut_H2 = QShortcut(QKeySequence("Alt+2"), self.noteArea)
        self.shortcut_H2.activated.connect(self.fbH2)
        self.shortcut_H3 = QShortcut(QKeySequence("Alt+3"), self.noteArea)
        self.shortcut_H3.activated.connect(self.fbH3)
        self.shortcut_H4 = QShortcut(QKeySequence("Alt+4"), self.noteArea)
        self.shortcut_H4.activated.connect(self.fbH4)
        self.shortcut_H5 = QShortcut(QKeySequence("Alt+5"), self.noteArea)
        self.shortcut_H5.activated.connect(self.fbH5)
        self.shortcut_H6 = QShortcut(QKeySequence("Alt+6"), self.noteArea)
        self.shortcut_H6.activated.connect(self.fbH6)

        self.shortcut_Bq = QShortcut(QKeySequence("Alt+7"), self.noteArea)
        self.shortcut_Bq.activated.connect(self.fbBq)
        self.shortcut_Pre = QShortcut(QKeySequence("Alt+8"), self.noteArea)
        self.shortcut_Pre.activated.connect(self.fbPre)
        self.shortcut_Addr = QShortcut(QKeySequence("Alt+9"), self.noteArea)
        self.shortcut_Addr.activated.connect(self.fbAddr)

    def _initTool(self):
        '''擺放工具列'''
        self.editToolbar.addWidget(self.headingBox)
        self.editToolbar.addWidget(self.fontsize)
        self.editToolbar.addAction(self.noformat)
        self.editToolbar.addAction(self.bold)
        self.editToolbar.addAction(self.italic)
        self.editToolbar.addAction(self.underline)
        self.editToolbar.addAction(self.strike)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.ordlist)
        self.editToolbar.addAction(self.bullist)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.inimg)
        self.editToolbar.addAction(self.inatt)
        self.editToolbar.addAction(self.inhtml)
        self.editToolbar.addAction(self.intemplate)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.indate)
        self.editToolbar.addAction(self.intime)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.link)
        self.editToolbar.addAction(self.unlink)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.outdent)
        self.editToolbar.addAction(self.indent)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.fgcolor)
        self.editToolbar.addAction(self.bgcolor)
        self.editToolbar.addSeparator()
        self.editToolbar.addAction(self.justifyLeft)
        self.editToolbar.addAction(self.justifyCenter)
        self.editToolbar.addAction(self.justifyRight)
        self.editToolbar.addAction(self.superscript)
        self.editToolbar.addAction(self.subscript)

#外部操作介面

    def inject(self, id = -1, htmlstr = '', notePath = '', cssPath = '', embPath = '', attPath = '', scrollpos = None):
        '''將html原始碼連同新的路徑組態一起注入
        id = 頁面的id，影響cache操作。
        htmlstr = 要讓editor顯示的內容
        cssPath = 外部css路徑
        embPath = 所有內嵌元素的基礎路徑，也是圖片插入功能的目標
        attPath = 附件插入功能的插入目的地
        scrollpos = 指示要將捲頁位置滾到哪裡'''
        #先確認現在頁面id和之前不同
        #if self.pageid == id: #如果用這個的話，變更版本紀錄功能就會失效……
        #    return False
        #檢查id有沒有在cache裡，有就用cache的內容。
        if self._useCache(id, htmlstr, cssPath):
            pass
        #cache裡沒有就用外面輸入的資料更新
        else:
            self._toCache() #先存起來……
            self._genNewPage() #產生新page
            #設定滾輪棒位置預讀
            self.initScrollPos = scrollpos
            self.noteArea.page().loadFinished.connect(self._initScrollTo)
            #寫入css內容
            self.noteArea.page().settings().setUserStyleSheetUrl(QUrl.fromLocalFile(cssPath))
            #寫入page內容並更新預設js
            QBasePath = QUrl.fromLocalFile(embPath + '/')
            self.noteArea.page().mainFrame().setHtml(htmlstr, QBasePath)
            self.imgResizeEnable_Refresh()
            self.noteArea.page().mainFrame().evaluateJavaScript(gadget.JS_imgSizeControl)
            self.pageid = id
            self.pagedir = notePath
            self.pageemb = embPath
            self.pageatt = attPath
        #更新findAll高亮
        self.findAll()
        #更新放大率
        self.textSizeMultiplier_Refresh()

    def extract(self):
        '''將已經修改好的資料抽出'''
        return {'htmlstr':self.noteArea.page().mainFrame().toHtml(),
                'plainstr':self.noteArea.page().mainFrame().toPlainText(),
                'id':self.pageid,
                'scrollpos':self.noteArea.page().mainFrame().scrollBarValue(Qt.Vertical)}

## 外部操作介面的底層 ##

    def _initScrollTo(self):
        '''當LoadFinish後滾頁到指定位置'''
        if self.initScrollPos != None:
            self.noteArea.page().mainFrame().setScrollBarValue(Qt.Vertical, self.initScrollPos)

    def _showLink(self, link, title, textContent):
        '''連結被滑過時觸發的事件'''
        window = QApplication.activeWindow()
        #如果不是mainwindow就避開，否則會出錯
        #找不到直接調用mainwindow的方法Q_Q
        if not isinstance(window, QMainWindow):
            return
        #看看有沒有db資料，沒有就直接連結
        if not self.db:
            window.statusBar().showMessage(link)
            return
        #判斷是不是lowiki連結，lowiki要另外處理
        qurl = QUrl(link)
        if qurl.scheme() == 'lowiki':
            uuid = qurl.path()
            getlen = 30
            answer = self.db.execute('select name,contentplain from notes where uuid = ?', (uuid,)).fetchone()
            msg = _('Note: [ {0} ] ➜ {1}')
            if answer:
                msg = msg.format(answer[0], answer[1][:getlen])
                if len(answer[1]) > getlen:
                     msg += _('...')
            else:
                msg = msg.format('✖' ,_('PAGE DISAPPEAR!    ☆Hint: Maybe you should remove this link...'))
            window.statusBar().showMessage(msg)
        else:
            window.statusBar().showMessage(link)

    def _useCache(self, id, htmlstr, cssPath):
        '''嘗試使用cache裡的note
        如果對應的id存在，套用該page並回傳true
        失敗不做任何事，回傳false'''
        #如果當前頁面就是指定id的頁面
        if self.pageid == id:
            return False
        if id in self.cache:
            #如果資料不同，表示已經被不知道誰從哪裡複寫過了。放棄cache
            if self.cache[id][0].mainFrame().toHtml() != htmlstr:
                self.cache.pop(id)
                return False
            #假設資料相同，進行替換作業，將目前頁面寫入cache
            self._toCache()
            #將指定id的頁面從cache裡面讀出來
            page, self.pageemb, self.pageatt = self.cache.pop(id)
            self.noteArea.page().setView(None)
            self.noteArea.setPage(page)
            self.pageid = id
            #更新css內容
            #先清掉先前的快取
            self.noteArea.page().settings().setUserStyleSheetUrl(QUrl())
            self.noteArea.page().settings().setUserStyleSheetUrl(QUrl.fromLocalFile(cssPath))
            #更新JS
            self.imgResizeEnable_Refresh()
            #更新字體
            self.fontRefresh()
            return True
        else:
            return False

    def _toCache(self):
        '''把當前頁面寫入cache'''
        if self.pageid == None:
            return
        self.cache[self.pageid] = (self.noteArea.page(), self.pageemb, self.pageatt)

    def _genNewPage(self):
        '''中斷先前頁面的聯結，產生一個空白的新頁面，並做好基本設定'''
        #打斷原先聯結 (這是因為QtWebKit的bug，不先打斷會不正常)
        self.noteArea.page().setView(None)
        #生成新頁面並聯結view
        self.noteArea.setPage(QWebPage())
        #設定其他相關
        webSetting = self.noteArea.page().settings()
        webSetting.setAttribute(QWebSettings.PluginsEnabled, True)
        webSetting.setAttribute(QWebSettings.JavascriptEnabled, True)
        webSetting.setAttribute(QWebSettings.JavaEnabled, True)
        webSetting.setAttribute(QWebSettings.JavascriptCanOpenWindows, True)
        webSetting.setAttribute(QWebSettings.JavascriptCanAccessClipboard, True)
        #webSetting.setAttribute(QWebSettings.AcceleratedCompositingEnabled, True)

        #webSetting.setAttribute(QWebSettings.LocalContentCanAccessFileUrls, True)
        #webSetting.setAttribute(QWebSettings.LocalContentCanAccessRemoteUrls, True)
        #字體與字型
        self.fontRefresh()
        #可編輯
        self.noteArea.page().setContentEditable(True)
        #點擊反應
        self.noteArea.page().setLinkDelegationPolicy(QWebPage.DelegateAllLinks)
        #連結被滑過後觸發到指定函式
        self.noteArea.page().linkHovered.connect(self._showLink)

#頁面狀態維護

    def fontRefresh(self):
        '''設定當前的Font格式'''
        webSetting = self.noteArea.page().settings()
        if self.FONT_ST:
            webSetting.setFontFamily(QWebSettings.StandardFont, self.FONT_ST)
        else:
            webSetting.resetFontFamily(QWebSettings.StandardFont)
        #webSetting.setFontFamily(QWebSettings.SerifFont, 'wqy-microhei')
        #webSetting.setFontFamily(QWebSettings.SansSerifFont, 'wqy-microhei')
        #webSetting.setFontFamily(QWebSettings.CursiveFont, 'wqy-microhei')
        #webSetting.setFontFamily(QWebSettings.FantasyFont, 'wqy-microhei')
        #webSetting.setFontFamily(QWebSettings.FixedFont, 'DroidSansMono')

    def imgResizeEnable_Refresh(self):
        if self.IMG_FORMAT_RESIZE_ENABLE:
            self.noteArea.page().mainFrame().evaluateJavaScript('lonoteImgResizeEnable = true')
        else:
            self.noteArea.page().mainFrame().evaluateJavaScript('lonoteImgResizeEnable = false')

    def textSizeMultiplier_Refresh(self):
        '''更新文字的放大率'''
        self.noteArea.setTextSizeMultiplier(self.TEXT_SIZE_MULTIPLIER)

#頁面搜尋功能

    def findbarShow(self):
        '''把findbar打開'''
        self.findbar.show()
        self.findEdit.setFocus()
        self.findEdit.selectAll()

    def findbarHide(self):
        '''把findbar關上'''
        self.findbar.hide()
        self.noteArea.setFocus()

    def findAll(self, string = ''):
        '''將noteArea中所有指定的字串全部標示出來'''
        self.noteArea.findText('', QWebPage.HighlightAllOccurrences)
        self.noteArea.findText(self.findEdit.text(), QWebPage.HighlightAllOccurrences)

    def findNext(self, string = ''):
        '''尋找noteArea中指定的字串'''
        self.noteArea.findText(self.findEdit.text(), QWebPage.FindWrapsAroundDocument)

    def findBefore(self):
        '''尋找noteArea中指定的字串'''
        self.noteArea.findText(self.findEdit.text(), QWebPage.FindWrapsAroundDocument | QWebPage.FindBackward)

#特殊處理

    def _linkClicked(self, qurl):
        '''聯結被K之後的行動……'''
        if qurl.scheme() == 'lowiki': #如果是lowiki，發出信號
            if self.db:
                ans = self.db.execute('select id from notes where uuid = ?', (qurl.path(),)).fetchone()
                if ans:
                    self.lowikiClickedSignal.emit(ans[0])
        else:
            QDesktopServices.openUrl(qurl)

#編輯功能

## 去除格式 ##

    def _noformat(self):
        '''去除格式'''
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("removeFormat");')
        #self.noteArea.page().mainFrame().evaluateJavaScript(gadget.JS_removeTag)

    def _unlink(self):
        '''移除連結'''
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("unlink");')

## 文字格式變更 ##

    @pyqtSlot()
    def _bold(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("bold");')

    @pyqtSlot()
    def _italic(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("italic");')

    @pyqtSlot()
    def _underline(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("underline");')

    @pyqtSlot()
    def _strike(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("strikethrough");')

    @pyqtSlot()
    def _superscript(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("superscript");')

    @pyqtSlot()
    def _subscript(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("subscript");')

## 段落格式變更 ##

#### formatblock變更 ####

    @pyqtSlot(int)
    def _headingChanged(self, index):
        if index == 0:
            pass
        elif index == 1:
            self.fbP()
        elif index == 2:
            self.fbH1()
        elif index == 3:
            self.fbH2()
        elif index == 4:
            self.fbH3()
        elif index == 5:
            self.fbH4()
        elif index == 6:
            self.fbH5()
        elif index == 7:
            self.fbH6()
        elif index == 8:
            self.fbBq()
        elif index == 9:
            self.fbPre()
        elif index == 10:
            self.fbAddr()
        self.headingBox.setCurrentIndex(0)
        self.noteArea.setFocus()

    def fbP(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false,"p");')

    def fbH1(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h1");')

    def fbH2(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h2");')

    def fbH3(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h3");')

    def fbH4(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h4");')

    def fbH5(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h5");')

    def fbH6(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "h6");')

    def fbBq(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "blockquote");')

    def fbPre(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false, "pre");')

    def fbAddr(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("formatblock", false,"address");')

#### 字體大小 ####

    @pyqtSlot(int)
    def _fontsizeChanged(self, index):
        if index == 0:
            pass
        else:
            self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("fontsize", false,"{0}");'.format(index))
        self.fontsize.setCurrentIndex(0)
        self.noteArea.setFocus()

#### 縮進與對齊 ####

    @pyqtSlot()
    def _justifyLeft(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("JustifyLeft");')

    @pyqtSlot()
    def _justifyCenter(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("JustifyCenter");')

    @pyqtSlot()
    def _justifyRight(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("JustifyRight");')

    @pyqtSlot()
    def _indent(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("Indent");')

    @pyqtSlot()
    def _outdent(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("Outdent");')

#### 列表 ####

    @pyqtSlot()
    def _ordlist(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("InsertOrderedList");')

    @pyqtSlot()
    def _bullist(self):
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("InsertUnorderedList");')

#### 顏色 ####

    @pyqtSlot()
    def _fgcolor(self):
        color = QColorDialog.getColor(QColor(), self, _('Please select a font color...'))
        if not color.isValid():
            return
        rgb = (color.red(), color.green(), color.blue())
        hexcolor = gadget.rgbToHTMLColor(rgb)
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("ForeColor",false,"{0}");'.format(hexcolor))

    @pyqtSlot()
    def _bgcolor(self):
        color = QColorDialog.getColor(QColor(), self, _('Please select a background color...'))
        if not color.isValid():
            return
        rgb = (color.red(), color.green(), color.blue())
        hexcolor = gadget.rgbToHTMLColor(rgb)
        self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("BackColor",false,"{0}");'.format(hexcolor))

#### 插入 ####

    @pyqtSlot(list)
    def _inimg(self, imgList = []):
        '''插入圖片'''
        def preview(imagePath):
            pix = QPixmap(imagePath)
            if not pix.isNull():
                if pix.width() > 500 or pix.height() > 500: #只縮小不放大
                    pix = pix.scaled(QSize(500,500), Qt.KeepAspectRatio, Qt.SmoothTransformation)
            perviewLabel.setPixmap(pix)

        #如果imgList為空，讓使用者選擇
        if not imgList:
            #建立檔案選擇窗口
            dialog = QFileDialog(self, Qt.Dialog)
            perviewLabel = QLabel(dialog)
            dialog.currentChanged.connect(preview)
            dialog.setNameFilter(_("Image (*.png *.gif *.jpg *.jpeg *.svg *.bmp)"))
            dialog.setFileMode(QFileDialog.ExistingFiles)
            dialog.layout().addWidget(perviewLabel,1,3)
            dialog.setViewMode(QFileDialog.Detail)
            dialog.setMinimumSize(QSize(800,700))
            dialog.layout().setSizeConstraint(QLayout.SetMinimumSize)
            if dialog.exec():
                imgList = dialog.selectedFiles()
                if imgList == []:
                    return
            else:
                return

        #imgList中確定有值了

        #複製檔案並以相對路徑寫入
        #先解算來源與目標路徑
        for imgFile in imgList:
            fileName = os.path.split(imgFile)[1]
            fulltPath = os.path.join(self.pageemb, fileName)
            #有相同檔名的檔案時，變更目標檔名再拷貝。
            if os.path.isfile(fulltPath): #確認沒有相同檔名的檔案
                counter = 1
                name, ext = os.path.splitext(fileName)
                while True:
                    newName = name + '_' + str(counter) + ext
                    tmptPath = os.path.join(self.pageemb, newName)
                    if not os.path.isfile(tmptPath):
                        fulltPath = tmptPath
                        fileName = newName
                        break
                    counter += 1
            #拷貝
            gadget.copy(imgFile, fulltPath)
            #如果插入時要縮放的話……
            if self.IMG_FORMAT_INSERTRESIZE_ENABLE:
                img = QImage(fulltPath)
                if img.width() > self.IMG_FORMAT_WIDTH: #縮小不放大
                    img.scaled(QSize(self.IMG_FORMAT_WIDTH,1), Qt.KeepAspectRatioByExpanding, Qt.SmoothTransformation).save(fulltPath, quality = self.IMG_FORMAT_INSERTRESIZE_QUALITY)

            #替換格式字段中的變數{$FILENAME}、{$FILEPATH}與{$IMAGE}
            finalInsertString = self.IMG_FORMAT.replace('{$IMAGE}', '<img src="{$FILEPATH}" width={$PIC_WIDTH}px defaultwidth={$PIC_WIDTH}>')
            finalInsertString = finalInsertString.replace('{$FILENAME}', fileName)
            finalInsertString = finalInsertString.replace('{$FILEPATH}', fileName)
            #如果圖片大小比寬度限制大，則套用寬度限制，否則使用原圖寬度
            imgWidth = QImage(fulltPath).width()
            if(imgWidth > self.IMG_FORMAT_WIDTH):
                finalInsertString = finalInsertString.replace('{$PIC_WIDTH}', str(self.IMG_FORMAT_WIDTH))
            else:
                finalInsertString = finalInsertString.replace('{$PIC_WIDTH}', str(imgWidth))

            #將內容插入頁面
            self.noteArea.page().mainFrame().evaluateJavaScript('''
                    var st = '{0}';
                    document.execCommand("insertHTML", false, st);'''.format(finalInsertString))

        #插入全部完成後執行一次JS，確保頁面更新。
        self.noteArea.page().mainFrame().evaluateJavaScript(gadget.JS_imgSizeControl)
        #更新lonoteImgResizeEnable
        self.imgResizeEnable_Refresh()

    def _inatt(self):
        '''插入附件'''
        filePathList = QFileDialog.getOpenFileNames(
                parent = self,
                caption = _('Select a file you want to insert...'),
                directory = os.getcwd(),
                filter = _('all (*)'));
        if filePathList: #選中後繼續
            #複製檔案並以相對路徑寫入
            for filePath in filePathList:
                fileName = os.path.split(filePath)[1]
                gadget.copy(filePath, os.path.join(self.pageatt, fileName))
        self.SIGNALattachAdd.emit()

    def _inhtml(self):
        '''插入原始碼'''
        source = QInputDialog.getText(
                self,
                _('Insert HTML source code'),
                _('Insert HTML cource code on your cursor position.\nThis may be a simple hyperlink, or a youtube video.'),
                mode = QLineEdit.Normal,
                text = '<hr>')
        if source[0] == '':
            return
        self.noteArea.page().mainFrame().evaluateJavaScript('''
                        var st = '{0}';
                        document.execCommand("insertHTML", false, st);'''.format(source[0]))

    def _indate(self):
        '''插入當前日期'''
        self.noteArea.page().mainFrame().evaluateJavaScript('''
                        var st = '{0}';
                        document.execCommand("insertHTML", false, st);'''.format(time.strftime(self.DATE_FORMAT)))

    def _intime(self):
        '''插入當前時間'''
        self.noteArea.page().mainFrame().evaluateJavaScript('''
                        var st = '{0}';
                        document.execCommand("insertHTML", false, st);'''.format(time.strftime(self.TIME_FORMAT)))

    def _link(self):
        '''建立超連結'''
        selText = self.noteArea.page().selectedText()
        if selText == '': #沒有選擇文字則跳出
            return
        strlist = selText.split('???', 1)

        #如果scheme為空，則表示沒有指定協定，嘗試建立lowiki連結
        if QUrl(strlist[0]).scheme() == '' and self.db:

            def insertLowikiLink(uuid, name):
                '''插入Lowiki連結'''
                url = QUrl()
                url.setScheme('lowiki')
                url.setPath(uuid)
                insStr = '<a href={0}>{1}</a>'.format(url.toString(), name)
                self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("insertHTML", false, "{0}");'.format(insStr))

            def createMenu(items):
                '''輸入一組item list
                item格式為：(uuid, contentplain, parentid, name)
                依照順序建立起一個QMenu並回傳'''
                menu = QMenu()
                for item in items:
                    #拼裝提示字串
                    getlen = 50 #參考文字長度
                    #拼裝節點名稱
                    actstr = _('➥Link to [ {0} ]: ').format(item[3])
                    #追加參考字並判斷末端要不要刪節號
                    if len(item[1]) <= getlen:
                        actstr += item[1]
                    else:
                        actstr += item[1][:getlen] + _('...')
                    #追加parent name
                    pName = '' #parent name
                    if item[2]:
                        pName = self.db.execute('select name from notes where id = ?', (item[2],)).fetchone()[0]
                    else:
                        pName = _('Cover page')
                    actstr += ' ' + _('(') + pName + _(')')

                    #建立Action
                    act = QAction(actstr, menu) #擷取前50字作為索引
                    act.triggered.connect(functools.partial(insertLowikiLink, item[0], showname))
                    menu.addAction(act)
                return menu

            #確認showname
            showname = None
            if len(strlist) == 2 and len(strlist[1]) > 0:
                showname = strlist[1]
            else:
                showname = strlist[0]

            #進行精確搜尋
            items = self.db.execute('select uuid, contentplain, parentid, name from notes where name = ?', (strlist[0],)).fetchall()
            #進行模糊比對
            fuzzyItems = self.db.execute('select uuid,contentplain, parentid, name from notes where name like ? and name <> ?', ('%' + strlist[0] + '%', strlist[0])).fetchall()

            #建立連結或彈出選單
            if len(items) == 0: #沒有任何精確對應的筆記頁
                #即使只有一個也要列出選單讓使用者確認，因為這是模糊的
                if len(fuzzyItems) >= 1:
                    menu = createMenu(fuzzyItems)
                    menu.exec_(QCursor.pos())
            elif len(items) == 1: #剛好有一個精確對應的筆記頁
                insertLowikiLink(items[0][0], showname)
            else: #有兩個以上可以精確對應的筆記頁
                menu = createMenu(items + fuzzyItems)
                menu.exec_(QCursor.pos())

        else: #如果scheme不為空，則用普通方式建立連結
            if len(strlist) == 2:
                insStr = '<a href={0}>{1}</a>'.format(strlist[0], strlist[1])
                self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("insertHTML", false, "{0}");'.format(insStr))
            else:
                self.noteArea.page().mainFrame().evaluateJavaScript('document.execCommand("createlink", false, "{0}");'.format(selText))

    def _intemplate(self):
        '''插入模版'''
        def locate(pattern, root=os.getcwd()):
            '''Locate all files matching supplied filename pattern in and below
            supplied root directory.'''
            for path, dirs, files in os.walk(os.path.abspath(root)):
                for filename in fnmatch.filter(files, pattern):
                    yield os.path.join(path, filename)

        #先尋找模版
        tempFilesDisplay = []
        #找出所有指定路徑下的模版檔案(.html)
        for file in locate('*.html' ,root = path.templateDir):
            tempFilesDisplay.append(os.path.splitext(os.path.relpath(file, path.templateDir))[0])

        #讓使用者選擇模板
        value = QInputDialog.getItem(
                self,
                _('Select a template you want'),
                _('Insert this template on your cursor position.\nYou can create your own template in LoNote\'s template folder.'),
                tempFilesDisplay,
                editable = False)
        if value[1] == False:
            return
        targetFile = os.path.join(path.templateDir, value[0]+'.html')
        insertString = ''
        with open(targetFile, encoding='utf-8') as File:
            insertString = File.read()

        #處理多行問題，在每行的末端加入java換行記號
        insertString = insertString.replace('\n','\\\n')

        self.noteArea.page().mainFrame().evaluateJavaScript('''
                        var st = '{0}';
                        document.execCommand("insertHTML", false, st);'''.format(insertString))
